"""Dataclasses for all structs in the ConstructorWithStructArgs contract.

DO NOT EDIT.  This file was generated by pypechain.  See documentation at
https://github.com/delvtech/pypechain """

# super() call methods are generic, while our version adds values & types
# pylint: disable=arguments-differ
# contracts have PascalCase names
# pylint: disable=invalid-name
# contracts control how many attributes and arguments we have in generated code
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments
# unable to determine which imports will be used in the generated code
# pylint: disable=unused-import
# we don't need else statement if the other conditionals all have return,
# but it's easier to generate
# pylint: disable=no-else-return
from __future__ import annotations

from dataclasses import dataclass


@dataclass
class Items:
    """Items struct."""

    thing: str
    yesOrNo: bool


@dataclass
class Config:
    """Config struct."""

    name: str
    items: Items



"""Export all types from generated files.

DO NOT EDIT.  This file was generated by pypechain.  See documentation at
https://github.com/delvtech/pypechain"""

from .ConstructorWithStructArgsContract import *
from .ConstructorWithStructArgsTypes import *
from .NoConstructorContract import *
from .ConstructorNoArgsContract import *
from .ConstructorWithArgsContract import *



"""Tests for rendering the init file."""
import os

from pypechain.utilities.templates import get_jinja_env

# using pytest fixtures necessitates this.
# pylint: disable=redefined-outer-name

current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


class TestRenderInit:
    """Tests rendering for the init file using snapshots"""

    def test_render_init(self, snapshot):
        """
        Tests the code gen init file using snapshots
        """

        env = get_jinja_env()
        init_template = env.get_template("init.py.jinja2")

        # TODO: add return types to function calls

        file_names = [
            "ContractFile1",
            "ContractFile2",
            "ContractFile3",
        ]

        init_code = init_template.render(file_names=file_names)

        snapshot.snapshot_dir = "snapshots"  # This line is optional.
        snapshot.assert_match(init_code, "expected_init.py")



"""Functions to render Python files from an abi usng a jinja2 template."""
from __future__ import annotations

import re
from pathlib import Path
from typing import Any, NamedTuple

from web3.types import ABI

from pypechain.utilities.abi import (
    get_abi_constructor,
    get_abi_items,
    get_input_names,
    get_input_names_and_types,
    get_input_types,
    get_output_names,
    get_output_names_and_types,
    get_output_types,
    get_structs_for_abi,
    is_abi_event,
    is_abi_function,
    load_abi_from_file,
)
from pypechain.utilities.format import capitalize_first_letter_only
from pypechain.utilities.templates import get_jinja_env
from pypechain.utilities.types import EventData, FunctionData, SignatureData, gather_matching_types


def render_contract_file(contract_name: str, abi_file_path: Path) -> str:
    """Returns the serialized code of the contract file to be generated.

    Arguments
    ---------
    contract_template : Template
        A jinja template containging types for all structs within an abi.
    abi_file_path : Path
        The path to the abi file to parse.

    Returns
    -------
    str
        A serialized python file.
    """

    # TODO: break this function up or bundle arguments to save on variables
    # pylint: disable=too-many-locals

    env = get_jinja_env()
    templates = get_templates_for_contract_file(env)

    abi, bytecode = load_abi_from_file(abi_file_path)
    function_datas, constructor_data = get_function_datas(abi)
    event_datas = get_event_datas(abi)

    has_bytecode = bool(bytecode)
    has_events = bool(len(event_datas.values()))
    # if any function has overloading
    has_overloading = any(function_data["has_overloading"] for function_data in function_datas.values())

    structs_for_abi = get_structs_for_abi(abi)
    structs_used = gather_matching_types(list(function_datas.values()), list(structs_for_abi.keys()))

    functions_block = templates.functions_template.render(
        abi=abi,
        contract_name=contract_name,
        functions=function_datas,
        # TODO: use this data to add a typed constructor
        constructor=constructor_data,
    )

    events_block = templates.events_template.render(
        contract_name=contract_name,
        events=event_datas,
    )

    abi_block = templates.abi_template.render(
        abi=abi,
        bytecode=bytecode,
        contract_name=contract_name,
    )

    contract_block = templates.contract_template.render(
        has_bytecode=has_bytecode,
        has_events=has_events,
        contract_name=contract_name,
        constructor=constructor_data,
        functions=function_datas,
    )

    # if any function has overloading
    has_overloading = any(function_data["has_overloading"] for function_data in function_datas.values())
    has_multiple_return_values = any(
        function_data["has_multiple_return_values"] for function_data in function_datas.values()
    )

    # Render the template
    return templates.base_template.render(
        contract_name=contract_name,
        structs_used=structs_used,
        structs_for_abi=structs_for_abi,
        has_overloading=has_overloading,
        has_multiple_return_values=has_multiple_return_values,
        has_bytecode=has_bytecode,
        has_events=has_events,
        functions_block=functions_block,
        events_block=events_block,
        abi_block=abi_block,
        contract_block=contract_block,
        # TODO: use this data to add a typed constructor
        # constructor_data=constructor_data,
    )


def get_has_multiple_return_signatures(signature_datas: list[SignatureData]) -> bool:
    """If there are multiple return signatures for a smart contract function, we'll need to overload
       the call() method.  This method compares the output types of all the signatures of a method.

    Parameters
    ----------
    signature_datas : list[SignatureData]
        a list of SignatureData's to compare.

    Returns
    -------
    bool
        If there are multiple return signatures or not.
    """
    lists_equal = True
    first_output_types: list[str] | None = None
    for signature_data in signature_datas:
        if first_output_types is None:
            first_output_types = signature_data["output_types"]
        else:
            lists_equal = all(
                output_types_to_compare[0] == output_types_to_compare[1]
                for output_types_to_compare in zip(first_output_types, signature_data["output_types"])
            )
            if not lists_equal:
                break

    return not lists_equal


def get_has_multiple_return_values(signature_datas: list[SignatureData]) -> bool:
    """If there are multiple return values for a smart contract function, we'll need to overload
       the call() method. This method compares the output types of all the values of a method.

    Parameters
    ----------
    signature_datas : list[SignatureData]
        a list of SignatureData's to compare.

    Returns
    -------
    bool
        If there are multiple return signatures or not.
    """
    for signature_data in signature_datas:
        if len(signature_data["outputs"]) > 1:
            return True
    return False


class ContractTemplates(NamedTuple):
    """Templates for the generated contract file."""

    base_template: Any
    functions_template: Any
    events_template: Any
    abi_template: Any
    contract_template: Any


def get_templates_for_contract_file(env):
    """Templates for the generated contract file."""
    return ContractTemplates(
        base_template=env.get_template("contract.py/base.py.jinja2"),
        functions_template=env.get_template("contract.py/functions.py.jinja2"),
        events_template=env.get_template("contract.py/events.py.jinja2"),
        abi_template=env.get_template("contract.py/abi.py.jinja2"),
        contract_template=env.get_template("contract.py/contract.py.jinja2"),
    )


class GetFunctionDatasReturnValue(NamedTuple):
    """Return value for get_function_datas"""

    function_datas: dict[str, FunctionData]
    constructor_data: SignatureData | None


def get_function_datas(abi: ABI) -> GetFunctionDatasReturnValue:
    """Gets the information needed for the generated Contract file.

    Arguments
    ---------
    abi : ABI
        An application boundary interface for smart contract in json format.

    Returns
    -------
    tuple[dict[str, FunctionData], SignatureData | None]
        A tuple where the first value is a dictionary of FunctionData's keyed by function name and
        the second value is SignatureData for the constructor.
    """

    # handle constructor
    abi_constructor = get_abi_constructor(abi)
    constructor_data: SignatureData | None = (
        {
            "input_names_and_types": get_input_names_and_types(abi_constructor),
            "input_names": get_input_names(abi_constructor),
            "input_types": get_input_types(abi_constructor),
            "outputs": get_output_names(abi_constructor),
            "output_types": get_output_names_and_types(abi_constructor),
        }
        if abi_constructor
        else None
    )

    # handle all other functions
    function_datas: dict[str, FunctionData] = {}
    for abi_function in get_abi_items(abi):
        if is_abi_function(abi_function):
            name = abi_function.get("name", "")
            name = re.sub(r"\W|^(?=\d)", "_", name)
            signature_data: SignatureData = {
                "input_names_and_types": get_input_names_and_types(abi_function),
                "input_names": get_input_names(abi_function),
                "input_types": get_input_types(abi_function),
                "outputs": get_output_names(abi_function),
                "output_types": get_output_types(abi_function),
            }

            function_data: FunctionData = {
                "name": name,
                "capitalized_name": capitalize_first_letter_only(name),
                "signature_datas": [signature_data],
                "has_overloading": False,
                "has_multiple_return_signatures": False,
                "has_multiple_return_values": False,
            }
            if not function_datas.get(name):
                function_datas[name] = function_data
                function_datas[name]["has_multiple_return_values"] = get_has_multiple_return_values([signature_data])
            else:
                signature_datas = function_datas[name]["signature_datas"]
                signature_datas.append(signature_data)
                function_datas[name]["has_overloading"] = len(signature_datas) > 1
                function_datas[name]["has_multiple_return_signatures"] = get_has_multiple_return_signatures(
                    signature_datas
                )
                function_datas[name]["has_multiple_return_values"] = get_has_multiple_return_values(signature_datas)
    return GetFunctionDatasReturnValue(function_datas, constructor_data)


def get_event_datas(abi: ABI) -> dict[str, EventData]:
    """Gets the event datas required for the events template.

    Arguments
    ---------
    abi : ABI
        An application boundary interface for smart contract in json format.

    Returns
    -------
    dict[str, EventData]
        A dictionary of EventData's keyed by event name.
    """
    event_datas: dict[str, EventData] = {}
    for abi_event in get_abi_items(abi):
        if is_abi_event(abi_event):
            name = abi_event.get("name", "")
            event_data: EventData = {
                "name": name,
                "capitalized_name": capitalize_first_letter_only(name),
            }
            event_datas[name] = event_data
    return event_datas



"""Export all types from generated files.

DO NOT EDIT.  This file was generated by pypechain.  See documentation at
https://github.com/delvtech/pypechain"""

from .ReturnTypesContract import *
from .ReturnTypesTypes import *



"""Functions to render Python files from an abi usng a jinja2 template."""

import os
from pathlib import Path

from pypechain.render.contract import render_contract_file
from pypechain.render.types import render_types_file
from pypechain.utilities.file import write_string_to_file
from pypechain.utilities.format import format_file


def render_files(abi_file_path: str, output_dir: str, line_length: int = 120) -> list[str]:
    """Processes a single JSON file to generate class and types files."""

    # get names
    file_path = Path(abi_file_path)
    output_path = Path(output_dir)
    filename = file_path.name
    contract_name = os.path.splitext(filename)[0].replace("-", "")
    contract_path = output_path.joinpath(f"{contract_name}")

    # render the code for each file
    rendered_contract_code = render_contract_file(contract_name, file_path)
    rendered_types_code = render_types_file(contract_name, file_path)

    # keep track of file names to return
    file_names: list[str] = []

    contract_file_path = Path(f"{contract_path}Contract.py")
    write_string_to_file(contract_file_path, rendered_contract_code)
    format_file(contract_file_path, line_length)
    file_names.append(f"{contract_name}Contract")

    if rendered_types_code:
        types_file_path = Path(f"{contract_path}Types.py")
        write_string_to_file(types_file_path, rendered_types_code)
        format_file(types_file_path, line_length)
        file_names.append(f"{contract_name}Types")

    return file_names



"""Tests for ABI utilities."""
import json
import os

from .abi import get_structs_for_abi

# using pytest fixtures necessitates this.
# pylint: disable=redefined-outer-name

current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


class TestStructs:
    """Tests pipeline from bots making trades to viewing the trades in the db"""

    def test_hyperdrive_structs(
        self,
    ):
        """Runs the entire pipeline and checks the database at the end.
        All arguments are fixtures.
        """

        json_file_path = os.path.join(project_root, "example/abis/Example.json")
        with open(json_file_path, "r", encoding="utf8") as file:
            data = json.load(file)

        structs = get_structs_for_abi(data["contracts"]["example/contracts/Example.sol:Example"]["abi"])

        actual = list(structs)
        expected = ["SimpleStruct", "InnerStruct", "NestedStruct"]
        assert actual == expected
        assert all(a == b for a, b in zip(actual, expected))



"""Tests for overloading methods."""
from __future__ import annotations

import os

import pytest
from web3.exceptions import Web3ValidationError

from pypechain.test.overloading.types import OverloadedMethodsContract

# using pytest fixtures necessitates this.
# pylint: disable=redefined-outer-name

current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


@pytest.mark.usefixtures("process_contracts")
class TestOverloading:
    """Tests pipeline from bots making trades to viewing the trades in the db"""

    def test_overloading(self, w3):
        """Runs the entire pipeline and checks the database at the end.
        All arguments are fixtures.
        """

        deployed_contract = OverloadedMethodsContract.deploy(w3=w3, account=w3.eth.accounts[0])

        s = "test string"
        x = 1
        y = 2

        result = deployed_contract.functions.doSomething(s).call()
        assert result == "test string"

        result = deployed_contract.functions.doSomething(x).call()
        assert result == 1 * 2

        result = deployed_contract.functions.doSomething(x, y).call()
        assert result == 1 + 2

        with pytest.raises(Web3ValidationError) as err:
            result = deployed_contract.functions.doSomething(x, y, s).call()  # type: ignore
        assert "Could not identify the intended function with name `doSomething`" in str(err.value)
